from decimal import Decimal
import json

from django.db import models
from django.db.models import Model, QuerySet
from django.core import serializers



class Json:
    """Json class for work with json format."""

    @classmethod
    def parse(cls, obj: any):
        """Converts a Python value to a Json serializable Python value."""
        if isinstance(obj, (str, int, float, bool)):
            return obj
        if obj in (None,):
            return obj
        if isinstance(obj, (list, tuple)):
            return [cls.parse(e) for e in obj]
        if isinstance(obj, dict):
            return {k:cls.parse(v) for k,v in obj.items()}
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, Model):
            return cls.parse(object_to_dict(obj))
        if isinstance(obj, QuerySet):
            return json.loads(serializers.serialize("json", obj))
        return str(obj)

def call(obj, *args, **kwargs):
    if callable(obj):
        try:
            return obj(*args, **kwargs)
        except (TypeError):
            return obj
    return obj
    
def _getattr(obj, name):
    names = name.split("__")
    for _name in names:
        obj = getattr(obj, _name)
    return obj

def object_to_dict(obj: Model, fields: list = None, exclude: list = None, 
extra_fields: list = None, parse_json: bool = False):
    """It takes an instance of a Django model and returns a dictionary with the 
    names of its fields and their values.
    
    Parameters:
        - obj (Model): Django model instance.

        - fields (list): Optional. A list of fields names or methods of obj.

        - exclude (list): Optional. A list of fields names to exclude.

        - extra_fields (list): Optional. List of extra fields names to show 
        with fields parameters.

        - parse_json (bool): default = False. If true, Json.parse (value) will 
        be executed at each value in the dictionary.
        
    Returns:
        {fieldname: value, ...}
        
    """
    
    if fields is None:
        exclude_types = (models.ManyToOneRel, models.ManyToManyRel)
        fields = [f.name for f in obj.__class__._meta.get_fields() 
            if not isinstance(f, exclude_types)]
    
    if extra_fields:
        fields.extend(extra_fields)
        
    if exclude != None:
        fields = [f for f in fields if not f in exclude]
    
    dic = dict()
    for name in fields:
        try:
            value = call(_getattr(obj, name))
        except (AttributeError) as e:
            raise AttributeError("Field '%s' does not belong to '%s'. %s" % 
                (name, obj.__class__._meta.model_name, e))
        
        if parse_json:
            value = Json.parse(value)
        dic[name] = value
        
    return dic
            
