from django.shortcuts import render
from django.http import Http404, HttpResponse, JsonResponse
from django.views.generic import View, DetailView, ListView
from django.views.generic.base import ContextMixin
from django.utils.translation import gettext as _

from tojson.utils import Json, object_to_dict



def json_error(field: str = "__all__", message: str = "", code: str = "", 
status: int = 404, error_dict: dict = None) -> dict:
    """returns a dictionary with the default structure for a json error.
    
    Parameters:
        field (str): field name or '__all__'.

        message (str): the error message.

        code (str): error code.

        status (int): response status code (403, 404, 500, ...).

        error_dict (dict): this is optional django form.errors. If set, the 
        field, message, and code fields will be ignored.
    """
    if error_dict:
        errors = {}
        for key in error_dict:
            errors[key] = [{
                "message": getattr(e, "message", str(e)),
                "code": getattr(e, "code", "")} 
                for e in error_dict[key]]
        return {"error": True, "errors": errors}
    errors = {field: [{"message": message, "code": code}]}
    return {"error": True, "errors": errors}

def json_error_response(field: str = "__all__", message: str = "", code: str = "", 
status: int = 404, error_dict: dict = None) -> JsonResponse:
    """Retorna JsonResponse(json_error(*args, **kwargs), status=status)."""
    error_dict = json_error(field, message, code, status, error_dict)
    return JsonResponse(error_dict, status=status)

class Json404(Exception):
    pass


class JsonView(ContextMixin, View):
    """Base View that returns a JsonResponse."""

    def get_context_data(self, **kwargs):
        return Json.parse(kwargs)

    def render_to_response(self, data, **response_kwargs):
        return JsonResponse(data=data, **response_kwargs)


class DetailJsonView(JsonView, DetailView):
    """
    Generic DetailView that returns a JsonResponse.

    Class parameters:
        - fields (list): Optional. List of fields names to show.

        - exclude (list): Optional. List of fields names to exclude.

        - extra_fields (list): Optional. List of extra fields names to show 
        with fields parameters.

        - All DetailView parameters.

        Note: You can set a field name or a method of model on (fields or extra_fields).
    """
    fields = None
    exclude = None
    extra_fields = None
    # random_url_kwargs: Si se indica un nombre, los usuarios podrán utilizar 
    # este nombre para pasarlo por URL como parámetro y obtener así un objeto 
    # de forma aleatoria. Ejemplo: https://www.unolet.com/.../?random=true
    random_url_kwarg = None
    
    def get(self, request, *args, **kwargs):
        data = self.get_context_data()

        try:
            self.object = self.get_object()
        except (Json404) as e:
            return json_error_response(message=str(e))

        if self.object:
            data["pk"] = self.object.pk
            data["model"] = "%s.%s" % (self.model._meta.app_label, self.model._meta.model_name)
            data["fields"] = object_to_dict(self.object, fields=self.fields, exclude=self.exclude, extra_fields=self.extra_fields, parse_json=True)

        return self.render_to_response(data)

    def get_object(self, queryset=None):
        """
        Return the object the view is displaying.
        Require `self.queryset` and a `pk` or `slug` argument in the URLconf.
        Subclasses can override this to return any object.
        """
        # Use a custom queryset if provided; this is required for subclasses
        # like DateDetailView
        if queryset is None:
            queryset = self.get_queryset()

        # Next, try looking up by primary key.
        pk = self.kwargs.get(self.pk_url_kwarg) or self.request.GET.get(self.pk_url_kwarg) or self.request.GET.get("id") # New
        slug = self.kwargs.get(self.slug_url_kwarg) or self.request.GET.get(self.slug_url_kwarg) # New
        if (pk is not None):
            try:
                queryset = queryset.filter(pk=pk)
            except (ValueError) as e:
                raise Json404(e)

        # Next, try looking up by slug.
        if (slug is not None) and (pk is None or self.query_pk_and_slug):
            slug_field = self.get_slug_field()
            queryset = queryset.filter(**{slug_field: slug})

        # If none of those are defined, it's an error.
        if (pk is None) and (slug is None):
            if self.random_url_kwarg != None:
                if self.request.GET.get(self.random_url_kwarg):
                    return queryset.order_by("?").first()
            raise Json404( # New.
                "Generic detail view %s must be called with either an object "
                "pk or a id or a slug in the URLconf or URL GET parameters." % self.__class__.__name__
            )
        try:
            # Get the single item from the filtered queryset
            obj = queryset.get()
        except queryset.model.DoesNotExist:
            raise Json404(_("No %(verbose_name)s found matching the query") %
                          {'verbose_name': queryset.model._meta.verbose_name})
        return obj

  
class ListJsonView(JsonView, ListView):
    """
    Generic ListView view that returns a JsonResponse.
    
    Class parameters:
        - fields (list): Optional. a list of fields names to get on queryset.
    """
    fields = []

    def get(self, request, *args, **kwargs):
        data = self.get_context_data()
        self.queryset = self.get_queryset()

        if self.paginate_by:
            paginator = self.get_paginator(self.queryset, self.paginate_by)
            page = paginator.get_page(self.request.GET.get("page"))
            object_list = page.object_list
            data["paginator"] = {
                "count": object_list.count(),
                "page_range": list(paginator.page_range),
                "per_page": paginator.per_page,
                "num_pages": paginator.num_pages,
                "orphans": paginator.orphans,
                "allow_empty_first_page": paginator.allow_empty_first_page,
                "has_next": page.has_next(),
                "has_other_pages": page.has_other_pages(),
                "has_previous": page.has_previous(),
                "number": page.number,
            }
        else:
            object_list = self.queryset

        data["model"] = "%s.%s" % (self.queryset.model._meta.app_label, self.queryset.model._meta.model_name)
        data["count"] = object_list.count()
        data["paginate_by"] = self.paginate_by
        data["list"] = Json.parse(object_list)
        return self.render_to_response(data)


